######## Includes all functions used to construct  minimally rigid graphs in R^3 and classify them up to the last Henneberg move


## 1. graph2int

# INPUT: a simple undirected graph
#OUTPUT: graph's integere representation

def graph2int(G):
    M= G.adjacency_matrix()
    n=M.nrows()
    L=''
    r=n-1
    c=n-1
    row_len=n-1
    while c>1:
        c+=-1
        if c<=row_len:
            row_len+=-1
            c=n-1
            r+=-1
        L=str(M[c,r])+L
    return Integer(int(L,2))

## 2.int2graph

# INPUT: an integer
#OUTPUT: the corresponding graph

def int2graph(N):
    L=Integer(N).binary()
    n=ceil((1+sqrt(1+8*len(L)))/2)
    #print N, (L), n
    adjacencyMatrix=matrix(n,n)
    j=len(L)
    r=n-1
    c=n-1
    row_len=n-1
    while j>0:
        #print adjacencyMatrix[r,c]
        c+=-1
        if c<=row_len:
            row_len+=-1
            c=n-1
            r+=-1
        j-=1
        adjacencyMatrix[r,c]=L[j]

    return Graph(adjacencyMatrix+adjacencyMatrix.transpose())


## 3.H1_3D_canonical

#needs int2graph/graph2int
#INPUT: list of n-vertex graphs as integers (a)
#OUTPUT: list of non-isomorphic n+1-vertex graphs constructed with an H1 move

def H1_3D_canonical(a):
    n=len(a)
    b_temp=[]
    test=int2graph(a[0])
    m=test.num_verts()
    print_out=(n/10).ceil()
    
    for g in [1..n]:
        "g-> previous graphs counter"
	if mod(g,print_out)==0:
	    print("{} graphs completed. {} temporal graphs".format(g, len(b_temp)))
        
	G=int2graph(a[g-1])
	
        for i in [0..m-3]:
            for j in [i+1..m-2]:
                for k in [j+1..m-1]:
                    
                    #"Choosing (n 3) vertices"             
                    
                    J=copy(G)
                    
                    #"3 edge addition"
                    
                    J.add_edge(i,m)
                    J.add_edge(j,m)
                    J.add_edge(k,m)
		    J1=J.canonical_label()
		    b_temp.append(graph2int(J1))

	
	
                    
                   
                        
    b=uniq(b_temp)                              
    print("H1 complete! {} new graphs ".format(len(b)))

    return b   



## 4.H2_3D_canonical


#needs int2graph/graph2int
#INPUT: list of n-vertex graphs as integers (a)
#OUTPUT: list of non-isomorphic n+1-vertex graphs constructed with an H2 move

def H2_3D_canonical(a):
    
    n=len(a)
    test=int2graph(a[0])
    m=test.num_verts()
    b_temp=[]
    E=test.num_edges()
    print_out=(n/10).ceil()
    for g in range(n):
	G=int2graph(a[g])
        #if mod(g,print_out)==0 and g>1:
	    #print("{} graphs completed. {} temporal graphs".format(g, len(b_temp)))

        for i in range(E):
            
            J1=copy(G)
            J2=copy(J1)
            
            "L: list of vertices not in the edge"
            L=range(m)
            L.remove(J1.edges()[i][0])
            L.remove(J1.edges()[i][1])
            
            J2.delete_edge(J2.edges()[i])
            J2.add_edge(J1.edges()[i][0],m)
            J2.add_edge(J1.edges()[i][1],m)
            
            for j in range(m-3):
		J3=copy(J2)
		J3.add_edge(L[j],m)

                for k in [j+1..m-3]:
			
			J4=copy(J3)
			J4.add_edge(L[k],m)
			J5=J4.canonical_label()
			b_temp.append(graph2int(J5))  

        
    b=uniq(b_temp)
    #print("H2 complete! {} new graphs ".format(len(b)))
	                       
    return b


## 5.H2only_3D_canonical

#needs int2graph/graph2int
#INPUT: list of n-vertex graphs as integers (a)
#OUTPUT: list of non-isomorphic n+1-vertex graphs constructed with an H2 move and not with an H1 #move (degree argument)

def H2only_3D_canonical(a):
    
    n=len(a)
    b_temp=[]
    test=int2graph(a[0])
    m=test.num_verts()
    E=test.num_edges()
    #print_out=(n/10).ceil()

    for g in range(n):
	G=int2graph(a[g])
        #if mod(g,print_out)==0 and g>1:
	    #print("{} graphs completed. {} temporal graphs".format(g, len(b_temp)))

	Deg=G.degree()
	Deg_h=G.degree_histogram()

	if Deg_h[3]<3:
		mn = min(Deg)
		
		G_sub=copy(G)
		
		if mn==3:
			idx=[i for i, j in enumerate(Deg) if j == mn]
			G_sub.delete_vertex(idx[0])
			if len(idx)>1:
				G_sub.delete_vertex(idx[1])

		E_h2=G_sub.num_edges()
		#m_h2=G_sub.num_verts()
	
		

       		for i in range(E_h2):
    			J1=copy(G_sub)
    			J2=copy(G)

    			#L: list of vertices not in the edge"
    			L=range(m)
    			L.remove(J1.edges()[i][0])
    			L.remove(J1.edges()[i][1])

    			J2.delete_edge(J1.edges()[i])
    			J2.add_edge(J1.edges()[i][0],m)
    			J2.add_edge(J1.edges()[i][1],m)

    			if Deg_h[3]==0:
        			for j in range(m-3):
            				J3=copy(J2)
            				J3.add_edge(L[j],m)
					
					for k in [j+1..m-3]:
						J4=copy(J3)
						J4.add_edge(L[k],m)
						J5=J4.canonical_label()
						b_temp.append(graph2int(J5)) 
	
    			elif Deg_h[3]==1:
        			J3=copy(J2)
        			J3.add_edge(idx[0],m)
				L.remove(idx[0])

				for k in range(m-3):
					J4=copy(J3)
					J4.add_edge(L[k],m)
					J5=J4.canonical_label()
					b_temp.append(graph2int(J5))

			else:
				J3=copy(J2)
				J3.add_edge(idx[0],m)
				J3.add_edge(idx[1],m)
				J4=J3.canonical_label()
				b_temp.append(graph2int(J4))
                

        
    b=uniq(b_temp)
    print("H2 complete! {} new graphs ".format(len(b)))
	                       
    return b

## 6.H3x_3D_canonical

#needs int2graph/graph2int
#INPUT: list of n-vertex graphs as integers (a)
#OUTPUT: list of non-isomorphic n+1-vertex graphs constructed with an H3x move

def H3x_3D_canonical(a):
    n=len(a)
    b_temp=[]
    test=int2graph(a[0])
    m=test.num_verts()
    E1=test.num_edges()
    print_out=(n/10).ceil()
    
    for g in range(n):
	G=int2graph(a[g])
	if mod(g,print_out)==0 and g>1:
	    print("{} graphs completed. {} temporal graphs".format(g, len(b_temp)))

        for i in range(E1-1):
            J1=copy(G)
            J2=copy(J1)

            "L1: list of vertices not in the first edge"
            L1=range(m)
            
            L1.remove(J1.edges()[i][0])
            L1.remove(J1.edges()[i][1])
            
            J2=J2.subgraph(L1)
            E2=J2.size()
            
            for j in range(i+1,E1):
                J3=copy(J1)

                J3.delete_edge(J3.edges()[i])

                J3.add_edge(J1.edges()[i][0],m)
                J3.add_edge(J1.edges()[i][1],m)


                J3.delete_edge(J3.edges()[j])
                
                J3.add_edge(J1.edges()[j][0],m)
                J3.add_edge(J1.edges()[j][1],m)
    
                L2=copy(L1)
                L2.remove(J1.edges()[j][0])
                L2.remove(J1.edges()[j][1])
                
                
                for k in range(len(L2)):
                    
                    J4=copy(J3)
                    J4.add_edge(L2[k],m)
                    J5=J4.canonical_label()
                    b_temp.append(graph2int(J5)) 
                    
                   
    b=uniq(b_temp)
    print("H3x complete! {} new graphs ".format(len(b)))
 
    return b



## 7.H3v_3D_canonical

#needs int2graph/graph2int
#INPUT: list of n-vertex graphs as integers (a)
#OUTPUT: list of non-isomorphic n+1-vertex graphs constructed with a v move in one graph 
#(Not two, as it is conjectured. This means that this function constructs a set that contains the one of the graphs constructed with an H3v move)

def H3v_3D_canonical(a):
    n=len(a)
    b_temp=[]
    #b_not_temp=[]
    print_out=(n/10).ceil()
    test=int2graph(a[0])
    m=test.num_verts()
    

    for g in range(n):
        "g-> previous graphs counter"
	if mod(g,print_out)==0 and g>0:
	    print("{} graphs completed. {} temporal graphs".format(g, len(b_temp)))
        
	G=int2graph(a[g])
	V=G.vertices()
	
        for i in range(m):
		
		L=copy(V)
		L.remove(V[i])

		V_n=G.neighbors(V[i])
		C_V_n=Combinations(V_n,2).list()
		

		for j in range(len(C_V_n)):
			J1=copy(G)
			J1.delete_edge(V[i],C_V_n[j][0])
			J1.delete_edge(V[i],C_V_n[j][1])

			L1=copy(L)
			L1.remove(C_V_n[j][0])
			L1.remove(C_V_n[j][1])
			
			J1.add_edge(V[i],V[m-1]+1)
			J1.add_edge(C_V_n[j][0],V[m-1]+1)
			J1.add_edge(C_V_n[j][1],V[m-1]+1)

			C2=Combinations(L1,2).list()

			for k in range(len(C2)):
				J2=copy(J1)
				J2.add_edge(C2[k][0],V[m-1]+1)
				J2.add_edge(C2[k][1],V[m-1]+1)
				
				if Laman_cond(J2,3)==1:
					
					b_temp.append(graph2int(J3))
				#else:
				#	J3=J2.canonical_label()
				#	b_not_temp.append(test_J)
				
			
			
            
                    
                    

	
	
                    
                   
                        
    b=uniq(b_temp)
    #b_not=uniq(b_not_temp)                             
    print("H3v complete! {} new graphs ".format(len(b)))

    return b   


## 8.Laman_cond

# INPUT: a graph, dimension of embedding-optonally
#OUTPUT: 1 if Laman condition holds for the given graph, 0 if Laman condition does not hold

def Laman_cond(G,flag_dim=0):
    V=G.num_verts()
    E=G.num_edges()
    
    if flag_dim>0:
        dim=flag_dim
    else:
        dim=1
        while (dim*V-(dim+1)*dim/2)<E:
            dim=dim+1
        
    
    L=dim*V-(dim+1)*dim/2
    
    flag=1

    if not L==E:
	flag=0
    
    k=V-1
    while flag==1 and k>dim+1:
	C = Combinations(range(V),k).list()
        i=0
	L_sub=dim*k-(dim+1)*dim/2
        while flag==1 and i<len(C):
		G_sub=G.subgraph(C[i])
		if G_sub.num_edges()>L_sub:
			flag=0
		i=i+1

	k=k-1

    if flag_dim>0:
        
        return flag
    else:
        return flag, dim
    
    
## 9.RigidityMatrix


#INPUT: graph, dimension of embedding, base_ring (default Q)
#OUTPUT: Rigidity matrix , Rank of RM, rigid/non rigid flag

def RigidityMatrix(G,d,base_ring=QQ):

    
    A=G.adjacency_matrix()
    V=G.order()
    E=G.size()
    X=Matrix(base_ring, V,d, lambda i,j: QQ.random_element(397,93))
    RM=Matrix(base_ring,d*V,E)

   
    k=0
    for i in range(V-1):
        for j in range(i+1,V):
            if A[i,j]==1:
                RM[d*i:d*i+d,k]=(X[i,:]-X[j,:]).transpose()
                RM[d*j:d*j+d,k]=(X[j,:]-X[i,:]).transpose()
                k=k+1
			
    Rank=RM.rank()
    flag='Non-Rigid'
    Laman=d*V-(d+1)*d/2
    
    if Laman==Rank:
        flag='Rigid'


    return RM, Rank,flag

## 10.remove_intersection_numbers

#removes intersection of numerical lists a and b from list a ('a-b'), lists with unique elements

def remove_intersection_numbers(b,a):
    a_new=copy(a)
    l_union=len(uniq(a+b))
    l_diff=l_union-len(b)
    

    i=0
    j=0
    a.sort()
    b.sort()
    l_b=len(b)

    while i<len(a) and len(a_new)>l_diff:
	flag=0
	
	while flag==0 and j<l_b:
		if a[i]==b[j]:
			a_new.remove(a[i])
			flag=1
			j=j+1
		elif a[i]<b[j]:
			flag=flag+1
			
		else:
			j=j+1
		
	i=i+1
    
    if len(b)==len(a) and a_new==[]:
	print('Same sets')

    elif len(a_new)==len(a):
	print('Disjoint sets')

    return a_new

